#include "bd.h"

#include <limits.h>
#include <setjmp.h>
#include <signal.h>

#include <memory>
#include <queue>
#include <set>
#include <unordered_set>

struct vector_string_hash {
  size_t operator()(const vector<string>& v) const {
    size_t h = 0;
    for (const string& s : v) {
      hash<string> f;
      h ^= f(s) + 0x9e3779b9 + (h<<6) + (h>>2);
    }
    return h;
  }
};

struct LambdaBFS {
  explicit LambdaBFS(const BD* bd)
    : orig_bd(bd) {}

  bool done(const BD* bd) const {
    return orig_bd->get(bd->x(), bd->y()) == '\\';
  }

  const BD* orig_bd;
};

struct NearLambdaBFS {
  explicit NearLambdaBFS(const BD* bd)
    : orig_bd(bd) {}

  bool done(const BD* bd) const {
    for (int dy = -2; dy <= 2; dy++) {
      int w = 2 - abs(dy);
      for (int dx = -w; dx <= w; dx++) {
        if (orig_bd->safeGet(bd->x()+dx, bd->y()+dy) == '\\')
          return true;
      }
    }
    return false;
  }

  const BD* orig_bd;
};

template<class Criterion>
static BD* searchByBFS(const BD* bd, const Criterion& criterion) {
  queue<BD*> q;
  q.push(new BD(*bd));

  set<pair<int, int> > seen;
  const string& cmds = "LRUD";

  BD* found = NULL;
  while (!q.empty()) {
    unique_ptr<BD> bd(q.front());
    q.pop();

    if (!seen.insert(make_pair(bd->x(), bd->y())).second)
      continue;

    for (char c : cmds) {
      unique_ptr<BD> nbd(new BD(*bd));
      nbd->tick(c);
      if (nbd->dead())
        continue;
      if (criterion.done(nbd.get())) {
        found = nbd.release();
        goto done;
      }
      q.push(nbd.release());
    }
  }
done:

  while (!q.empty()) {
    delete q.front();
    q.pop();
  }

  return found;
}

static BD* searchLambdaBFS(const BD* bd) {
  NearLambdaBFS bfs(bd);
  if (bfs.done(bd)) {
    LambdaBFS bfs(bd);
    return searchByBFS(bd, bfs);
  } else {
    return searchByBFS(bd, bfs);
  }
}

struct GoalBFS {
  explicit GoalBFS(BD* bd)
    : gx(0), gy(0) {
    if (!bd->makeGoal(&gx, &gy))
      assert(false);
  }

  bool done(const BD* bd) const {
    return bd->x() == gx && bd->y() == gy;
  }

  int gx, gy;
};

static BD* searchGoalBFS(BD* bd) {
  GoalBFS bfs(bd);
  return searchByBFS(bd, bfs);
}

bool isGoalConnected(const BD* orig_bd) {
  unique_ptr<BD> bd(new BD(*orig_bd));
  bd.reset(searchGoalBFS(bd.get()));
  return bd.get();
}

struct FreedomBFS {
  bool done(const BD* bd) const {
    return bd->goal() || bd->seq().size() >= 5;
  }
};

bool hasSomeFreedom(const BD* orig_bd) {
  unique_ptr<BD> bd(new BD(*orig_bd));
  bd->reset();
  FreedomBFS bfs;
  bd.reset(searchByBFS(bd.get(), bfs));
  return bd.get();
}

int getExpectedScoreByBFS(const BD* orig_bd, int cnt) {
  int best_score = orig_bd->score();
  unique_ptr<BD> bd(new BD(*orig_bd));
  while (cnt-- && bd->lambda() != bd->num_lambda()) {
    BD* nbd = searchLambdaBFS(bd.get());
    if (nbd) {
      bd.reset(nbd);
      best_score = max(bd->getFinishScore(), best_score);
    } else {
      return best_score;
    }
  }

  unique_ptr<BD> nbd(searchGoalBFS(bd.get()));
  if (nbd) {
    best_score = max(best_score, nbd->getFinishScore());
  }
  return best_score;
}

class SolverBase {
public:
  SolverBase() {}

  const BD* best_bd() const { return best_bd_.get(); }

  BD* releaseBestBD() { return best_bd_.release(); }

protected:
  unique_ptr<BD> best_bd_;
  unordered_set<vector<string>, vector_string_hash> seen_;
};

class GotoSolver : public SolverBase {
public:
  struct Better {
    Better(int gx0, int gy0)
      : gx(gx0), gy(gy0) {}
    bool operator()(BD* l, BD* r) const {
      int ld = l->getDistance(gx, gy);
      int rd = r->getDistance(gx, gy);
      if (ld > rd)
        return true;
      if (ld < rd)
        return false;
      return l->score() < r->score();
    }

    int gx, gy;
  };

  GotoSolver(BD* bd, int gx, int gy, bool stop_on_lambda = false)
    : cmds_("LRUDW"),
      q_(Better(gx, gy)),
      gx_(gx),
      gy_(gy),
      orig_bd_(bd),
      orig_lambda_(bd->lambda()),
      stop_on_lambda_(stop_on_lambda) {
    distance_ = bd->getDistance(gx, gy);
    q_.push(bd);
  }

  ~GotoSolver() {
    while (!q_.empty()) {
      delete q_.top();
      q_.pop();
    }
  }

  void run() {
    if (stop_on_lambda_) {
      best_bd_.reset(searchLambdaBFS(orig_bd_));
      if (best_bd_) {
        LOGF("by bfs: %s\n", best_bd_->seq().c_str());
        return;
      }
    }

    while (!q_.empty()) {
      unique_ptr<BD> bd(q_.top());
      q_.pop();

      if (stop_on_lambda_ && bd->lambda() > orig_lambda_) {
        best_bd_.reset(bd.release());
        return;
      }

      if ((int)bd->seq().size() + bd->getDistance(gx_, gy_) > distance_ * 3)
        continue;

      if (!seen_.insert(bd->mine()).second) {
        continue;
      }

      if (bd->x() == gx_ && bd->y() == gy_) {
        best_bd_.reset(bd.release());
        break;
      }

      for (char c : cmds_) {
        unique_ptr<BD> nbd(new BD(*bd));
        nbd->tick(c);
        if (nbd->dead())
          continue;
        q_.push(nbd.release());
      }
    }

    if (seen_.size() <= 1)
      best_bd_.reset();
  }

private:
  const string cmds_;
  priority_queue<BD*, vector<BD*>, Better> q_;
  int gx_, gy_;
  int distance_;
  BD* orig_bd_;
  int orig_lambda_;
  bool stop_on_lambda_;
};

class RangeSolver : public SolverBase {
public:
  explicit RangeSolver(BD* bd)
    : cmds_("LRUDW"),
      best_score_(make_pair(INT_MIN, INT_MIN)),
      orig_lambda_(bd->lambda()),
      min_x_(0),
      min_y_(0),
      max_x_(bd->w() - 1),
      max_y_(bd->h() - 1) {
    q_.push(bd);
    num_lambda_ = bd->num_lambda();
  }

  RangeSolver(BD* bd, int min_x, int min_y, int max_x, int max_y)
    : cmds_("LRUDW"),
      best_score_(make_pair(INT_MIN, INT_MIN)),
      min_x_(min_x),
      min_y_(min_y),
      max_x_(max_x),
      max_y_(max_y) {
    q_.push(bd);
    assert(min_x >= 0);
    assert(min_y >= 0);
    assert(max_x < bd->w());
    assert(max_y < bd->h());
    assert(min_x < max_x);
    assert(min_y < max_y);
    num_lambda_ = 0;
    for (int y = min_y; y < max_y; y++) {
      for (int x = min_x; x < max_x; x++) {
        if (bd->mine()[y][x] == '\\')
          num_lambda_++;
      }
    }
  }

  ~RangeSolver() {
    while (!q_.empty()) {
      delete q_.top();
      q_.pop();
    }
  }

  void showRange() const {
    LOGF("x=%d-%d y=%d-%d\n",
         min_x_, max_x_, min_y_, max_y_);
  }

  int num_lambda() const { return num_lambda_; }

  void run() {
    int comp_best_score = INT_MIN;

    while (!q_.empty()) {
      unique_ptr<BD> bd(q_.top());
      q_.pop();

      //bd->show();
      if (comp_best_score > bd->score() /* + bd->getLeftLambda() * 25 */) {
        continue;
      }

      if (bd->x() < min_x_ || bd->x() > max_x_ ||
          bd->y() < min_y_ || bd->y() > max_y_) {
        continue;
      }

      //showRange();
      //LOGF("%d %d\n", bd->x(), bd->y());

      if (!seen_.insert(bd->mine()).second) {
        continue;
      }

      for (char c : cmds_) {
        unique_ptr<BD> nbd(new BD(*bd));
        nbd->tick(c);
        if (nbd->dead()) {
          continue;
        }

        int score = nbd->score();
        //LOGF("score=%d=>%d\n", score, getExpectedScoreByBFS(nbd.get()));
/*
        if (best_score_.first < score + 25) {
          int expected_score = getExpectedScoreByBFS(nbd.get(), 3);

          if (best_score_.first + best_score_.second <
              score + expected_score) {
            best_score_ = make_pair(score, expected_score);
*/

/*
        if (best_score_.first < score || !best_score_.second) {
          int connected = isGoalConnected(nbd.get());
          if (best_score_.second < connected ||
              (best_score_.second == connected && best_score_.first < score)) {
            best_score_.first = score;
            best_score_.second = connected;
*/

        if (best_score_.first < score || !best_score_.second) {
          int freedom = hasSomeFreedom(nbd.get());
          if (best_score_.second < freedom ||
              (best_score_.second == freedom && best_score_.first < score)) {
            best_score_.first = score;
            best_score_.second = freedom;

/*
        if (best_score_.first < score) {
          if (1) {
            best_score_.first = score;
*/
            LOGF("score=%d seq=%s second=%d\n",
                 nbd->score(), nbd->seq().c_str(), best_score_.second);
            best_bd_.reset(new BD(*nbd));
          }
        }

        if (nbd->lambda() == num_lambda_) {
          if (comp_best_score < best_score_.first)
            comp_best_score = best_score_.first;
        } else {
          q_.push(nbd.release());
        }
      }
    }

    if (best_bd_) {
/*
  LOGF("best_bd_size=%d %d %d\n",
       best_bd_->seq().size(), best_bd_->lambda(), orig_lambda_);
*/
      if (best_bd_->seq().size() <= 1 && best_bd_->lambda() <= orig_lambda_)
        best_bd_.reset();
    }
  }

private:
  priority_queue<BD*, vector<BD*>, BD::Better> q_;
  const string cmds_;
  pair<int, int> best_score_;
  int orig_lambda_;
  int num_lambda_;
  int min_x_, min_y_, max_x_, max_y_;
};

class RootSolver {
public:
  explicit RootSolver(BD* bd)
    : bd_(bd) {
    best_seq_ = strdup("A");
    best_score_ = 0;
  }

  ~RootSolver() {
    free((char*)best_seq_);
  }

  void run() {
    for (;;) {
      bd_->reset();

      if (bd_->num_lambda() == 0) {
        int gx, gy;
        bd_->getGoal(&gx, &gy);

        GotoSolver solver(new BD(*bd_), gx, gy);
        solver.run();

        BD* last = solver.releaseBestBD();
        if (last) {
          bd_.reset(last);
          addSeq();
        } else {
          bd_->tick('A');
          addSeq();
        }
        return;
      }

      BD* range = stepRange();
      if (range) {
        bd_.reset(range);
        addSeq();
      } else {
        BD* go = stepGoto();
        // give up
        if (!go) {
          bd_->tick('A');
          addSeq();
          return;
        }
        bd_.reset(go);
        bd_->show();
        addSeq();
      }
    }
  }

  const string& seq() const { return seq_; }
  const char* best_seq() const { return best_seq_; }
  int best_score() const { return best_score_; }

  const BD* bd() { return bd_.get(); }

private:
  void addSeq() {
    seq_ += bd_->seq();
    int score = bd_->getFinishScore();
    LOGF("addSeq: %d %d %d\n", best_score_, score, bd_->state());
    if (best_score_ <= score) {
      best_score_ = score;
      char* seq = (char*)malloc(seq_.size() + 1);
      strcpy(seq, seq_.data());
      char* old_seq = (char*)best_seq_;
      // Need an atomic update.
      best_seq_ = seq;
      free(old_seq);
    }
  }

  BD* stepRange() {
    LOGF("range solver\n");
    LOGF("xy=%d,%d\n", bd_->x(), bd_->y());
    bd_->show();
    RangeSolver* best_solver = NULL;
    int best_score = INT_MIN;

    int x = bd_->x();
    int y = bd_->y();
    for (int min_y = y - H_; min_y <= y; min_y++) {
      if (min_y < 0 || min_y + H_ >= bd_->h())
        continue;
      for (int min_x = x - W_; min_x <= x; min_x++) {
        if (min_x < 0 || min_x + W_ >= bd_->w())
          continue;

        int max_x = min_x + W_;
        int max_y = min_y + H_;
        unique_ptr<RangeSolver> solver(new RangeSolver(
                                         new BD(*bd_),
                                         min_x, min_y, max_x, max_y));
        if (!solver->num_lambda()) {
          continue;
        }
        solver->showRange();
        solver->run();

        if (!solver->best_bd()) {
          continue;
        }

        if (best_score < solver->best_bd()->score()) {
          best_score = solver->best_bd()->score();
          best_solver = solver.release();
        }
      }
    }

    if (!best_solver) {
      return NULL;
    }

    best_solver->showRange();
    best_solver->best_bd()->show();
    BD* r = best_solver->releaseBestBD();
    delete best_solver;
    return r;
  }

  BD* stepGoto() {
    int best_d = INT_MAX;
    int best_x = INT_MIN;
    int best_y = INT_MIN;
    for (int y = 0; y < bd_->h(); y++) {
      for (int x = 0; x < bd_->w(y); x++) {
        if (bd_->get(x, y) != '\\')
          continue;

        int d = abs(x - bd_->x()) + abs(y - bd_->y());
        if (best_d > d) {
          best_d = d;
          best_x = x;
          best_y = y;
        }
      }
    }

    LOGF("goto solver %d %d\n", best_x, best_y);

    if (best_d == INT_MAX) {
      return NULL;
    }

    GotoSolver solver(new BD(*bd_), best_x, best_y, true);
    solver.run();
    if (!solver.best_bd()) {
      LOGF("failed to goto %d %d\n", best_x, best_y);
      return NULL;
    }

    return solver.releaseBestBD();
  }

  unique_ptr<BD> bd_;
  string seq_;
  const char* best_seq_;
  int best_score_;
  static const int W_ = 5;
  static const int H_ = 5;
};

static sigjmp_buf g_jmp_buf;
static void handleSignal(int) {
  // unsafe!
  siglongjmp(g_jmp_buf, 1);
}

int main(int argc, char* argv[]) {
  FILE* map_fp = stdin;
  if (argc >= 2) {
    map_fp = fopen(argv[1], "rb");
    assert(map_fp);
  }

  BD* bd = new BD(map_fp);
  fclose(map_fp);

/*
  Solver solver(bd);
  solver.run();
*/
  RootSolver solver(bd);

  if (sigsetjmp(g_jmp_buf, 1)) {
    // Maybe OK because we don't malloc
    write(1, solver.best_seq(), strlen(solver.best_seq()));
    write(1, "A\n", 2);
    _exit(0);
  }

  signal(SIGINT, &handleSignal);
  signal(SIGSEGV, &handleSignal);
  solver.run();
  LOGF("score=%d\n", solver.best_score());
  printf("%sA\n", solver.best_seq());

#if 0
  const BD* best_bd = solver.best_bd();

  LOGF("score=%d\n", best_bd->score());
  printf("%s\n", best_bd->seq().c_str());
#endif
}
